@@ -58,7 +58,6 @@
 //! ```
 
 #![no_std]
-#![cfg_attr(feature = "nightly", feature(negative_impls, auto_traits))]
 
 #[cfg(feature = "hashbrown")]
 extern crate hashbrown;
@@ -74,7 +73,7 @@ use core::iter::FusedIterator;
 use core::marker::PhantomData;
 use core::mem;
 use core::num::NonZeroUsize;
-use core::ptr;
+use core::ptr::{self, NonNull};
 use core::usize;
 
 #[cfg(any(test, not(feature = "hashbrown")))]
@@ -88,8 +87,7 @@ use std::collections::HashMap;
 extern crate alloc;
 
 // Struct used to hold a reference to a key
-#[doc(hidden)]
-pub struct KeyRef<K> {
+struct KeyRef<K> {
     k: *const K,
 }
 
@@ -107,49 +105,40 @@ impl<K: PartialEq> PartialEq for KeyRef<K> {
 
 impl<K: Eq> Eq for KeyRef<K> {}
 
-#[cfg(feature = "nightly")]
-#[doc(hidden)]
-pub auto trait NotKeyRef {}
+// This type exists to allow a "blanket" Borrow impl for KeyRef without conflicting with the
+//  stdlib blanket impl
+#[repr(transparent)]
+struct KeyWrapper<K: ?Sized>(K);
 
-#[cfg(feature = "nightly")]
-impl<K> !NotKeyRef for KeyRef<K> {}
+impl<K: ?Sized> KeyWrapper<K> {
+    fn from_ref(key: &K) -> &Self {
+        // safety: KeyWrapper is transparent, so casting the ref like this is allowable
+        unsafe { &*(key as *const K as *const KeyWrapper<K>) }
+    }
+}
 
-#[cfg(feature = "nightly")]
-impl<K, D> Borrow<D> for KeyRef<K>
+impl<K: ?Sized + Hash> Hash for KeyWrapper<K> {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        self.0.hash(state)
+    }
+}
+
+impl<K: ?Sized + PartialEq> PartialEq for KeyWrapper<K> {
+    fn eq(&self, other: &Self) -> bool {
+        self.0.eq(&other.0)
+    }
+}
+
+impl<K: ?Sized + Eq> Eq for KeyWrapper<K> {}
+
+impl<K, Q> Borrow<KeyWrapper<Q>> for KeyRef<K>
 where
-    K: Borrow<D>,
-    D: NotKeyRef + ?Sized,
+    K: Borrow<Q>,
+    Q: ?Sized,
 {
-    fn borrow(&self) -> &D {
-        unsafe { &*self.k }.borrow()
-    }
-}
-
-#[cfg(not(feature = "nightly"))]
-impl<K> Borrow<K> for KeyRef<K> {
-    fn borrow(&self) -> &K {
-        unsafe { &*self.k }
-    }
-}
-
-#[cfg(not(feature = "nightly"))]
-impl Borrow<str> for KeyRef<alloc::string::String> {
-    fn borrow(&self) -> &str {
-        unsafe { &*self.k }
-    }
-}
-
-#[cfg(not(feature = "nightly"))]
-impl<T: ?Sized> Borrow<T> for KeyRef<Box<T>> {
-    fn borrow(&self) -> &T {
-        unsafe { &*self.k }
-    }
-}
-
-#[cfg(not(feature = "nightly"))]
-impl<T> Borrow<[T]> for KeyRef<alloc::vec::Vec<T>> {
-    fn borrow(&self) -> &[T] {
-        unsafe { &*self.k }
+    fn borrow(&self) -> &KeyWrapper<Q> {
+        let key = unsafe { &*self.k }.borrow();
+        KeyWrapper::from_ref(key)
     }
 }
 
@@ -189,7 +178,7 @@ pub type DefaultHasher = std::collections::hash_map::RandomState;
 
 /// An LRU Cache
 pub struct LruCache<K, V, S = DefaultHasher> {
-    map: HashMap<KeyRef<K>, Box<LruEntry<K, V>>, S>,
+    map: HashMap<KeyRef<K>, NonNull<LruEntry<K, V>>, S>,
     cap: NonZeroUsize,
 
     // head and tail are sigil nodes to facilitate inserting entries
@@ -266,7 +255,7 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// Creates a new LRU Cache with the given capacity.
     fn construct(
         cap: NonZeroUsize,
-        map: HashMap<KeyRef<K>, Box<LruEntry<K, V>>, S>,
+        map: HashMap<KeyRef<K>, NonNull<LruEntry<K, V>>, S>,
     ) -> LruCache<K, V, S> {
         // NB: The compiler warns that cache does not need to be marked as mutable if we
         // declare it as such since we only mutate it inside the unsafe block.
@@ -342,19 +331,23 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
 
         match node_ref {
             Some(node_ref) => {
-                let node_ptr: *mut LruEntry<K, V> = &mut **node_ref;
-
                 // if the key is already in the cache just update its value and move it to the
                 // front of the list
-                unsafe { mem::swap(&mut v, &mut (*(*node_ptr).val.as_mut_ptr()) as &mut V) }
+                let node_ptr: *mut LruEntry<K, V> = node_ref.as_ptr();
+
+                // gets a reference to the node to perform a swap and drops it right after
+                let node_ref = unsafe { &mut (*(*node_ptr).val.as_mut_ptr()) };
+                mem::swap(&mut v, node_ref);
+                let _ = node_ref;
+
                 self.detach(node_ptr);
                 self.attach(node_ptr);
                 Some((k, v))
             }
             None => {
-                let (replaced, mut node) = self.replace_or_create_node(k, v);
+                let (replaced, node) = self.replace_or_create_node(k, v);
+                let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
-                let node_ptr: *mut LruEntry<K, V> = &mut *node;
                 self.attach(node_ptr);
 
                 let keyref = unsafe { (*node_ptr).key.as_ptr() };
@@ -368,27 +361,32 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     // Used internally to swap out a node if the cache is full or to create a new node if space
     // is available. Shared between `put`, `push`, `get_or_insert`, and `get_or_insert_mut`.
     #[allow(clippy::type_complexity)]
-    fn replace_or_create_node(&mut self, k: K, v: V) -> (Option<(K, V)>, Box<LruEntry<K, V>>) {
+    fn replace_or_create_node(&mut self, k: K, v: V) -> (Option<(K, V)>, NonNull<LruEntry<K, V>>) {
         if self.len() == self.cap().get() {
             // if the cache is full, remove the last entry so we can use it for the new key
             let old_key = KeyRef {
                 k: unsafe { &(*(*(*self.tail).prev).key.as_ptr()) },
             };
-            let mut old_node = self.map.remove(&old_key).unwrap();
+            let old_node = self.map.remove(&old_key).unwrap();
+            let node_ptr: *mut LruEntry<K, V> = old_node.as_ptr();
 
             // read out the node's old key and value and then replace it
-            let replaced = unsafe { (old_node.key.assume_init(), old_node.val.assume_init()) };
+            let replaced = unsafe {
+                (
+                    mem::replace(&mut (*node_ptr).key, mem::MaybeUninit::new(k)).assume_init(),
+                    mem::replace(&mut (*node_ptr).val, mem::MaybeUninit::new(v)).assume_init(),
+                )
+            };
 
-            old_node.key = mem::MaybeUninit::new(k);
-            old_node.val = mem::MaybeUninit::new(v);
-
-            let node_ptr: *mut LruEntry<K, V> = &mut *old_node;
             self.detach(node_ptr);
 
             (Some(replaced), old_node)
         } else {
             // if the cache is not full allocate a new LruEntry
-            (None, Box::new(LruEntry::new(k, v)))
+            // Safety: We allocate, turn into raw, and get NonNull all in one step.
+            (None, unsafe {
+                NonNull::new_unchecked(Box::into_raw(Box::new(LruEntry::new(k, v))))
+            })
         }
     }
 
@@ -413,16 +411,16 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn get<'a, Q>(&'a mut self, k: &Q) -> Option<&'a V>
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        if let Some(node) = self.map.get_mut(k) {
-            let node_ptr: *mut LruEntry<K, V> = &mut **node;
+        if let Some(node) = self.map.get_mut(KeyWrapper::from_ref(k)) {
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
             self.detach(node_ptr);
             self.attach(node_ptr);
 
-            Some(unsafe { &(*(*node_ptr).val.as_ptr()) as &V })
+            Some(unsafe { &*(*node_ptr).val.as_ptr() })
         } else {
             None
         }
@@ -449,16 +447,16 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn get_mut<'a, Q>(&'a mut self, k: &Q) -> Option<&'a mut V>
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        if let Some(node) = self.map.get_mut(k) {
-            let node_ptr: *mut LruEntry<K, V> = &mut **node;
+        if let Some(node) = self.map.get_mut(KeyWrapper::from_ref(k)) {
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
             self.detach(node_ptr);
             self.attach(node_ptr);
 
-            Some(unsafe { &mut (*(*node_ptr).val.as_mut_ptr()) as &mut V })
+            Some(unsafe { &mut *(*node_ptr).val.as_mut_ptr() })
         } else {
             None
         }
@@ -491,22 +489,77 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
         F: FnOnce() -> V,
     {
         if let Some(node) = self.map.get_mut(&KeyRef { k: &k }) {
-            let node_ptr: *mut LruEntry<K, V> = &mut **node;
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
             self.detach(node_ptr);
             self.attach(node_ptr);
 
-            unsafe { &(*(*node_ptr).val.as_ptr()) as &V }
+            unsafe { &*(*node_ptr).val.as_ptr() }
         } else {
             let v = f();
-            let (_, mut node) = self.replace_or_create_node(k, v);
+            let (_, node) = self.replace_or_create_node(k, v);
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
-            let node_ptr: *mut LruEntry<K, V> = &mut *node;
             self.attach(node_ptr);
 
             let keyref = unsafe { (*node_ptr).key.as_ptr() };
             self.map.insert(KeyRef { k: keyref }, node);
-            unsafe { &(*(*node_ptr).val.as_ptr()) as &V }
+            unsafe { &*(*node_ptr).val.as_ptr() }
+        }
+    }
+
+    /// Returns a reference to the value of the key in the cache if it is
+    /// present in the cache and moves the key to the head of the LRU list.
+    /// If the key does not exist the provided `FnOnce` is used to populate
+    /// the list and a reference is returned. If `FnOnce` returns `Err`,
+    /// returns the `Err`.
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use lru::LruCache;
+    /// use std::num::NonZeroUsize;
+    /// let mut cache = LruCache::new(NonZeroUsize::new(2).unwrap());
+    ///
+    /// cache.put(1, "a");
+    /// cache.put(2, "b");
+    /// cache.put(2, "c");
+    /// cache.put(3, "d");
+    ///
+    /// let f = ||->Result<&str, String> {Err("failed".to_owned())};
+    /// let a = ||->Result<&str, String> {Ok("a")};
+    /// let b = ||->Result<&str, String> {Ok("b")};
+    /// assert_eq!(cache.try_get_or_insert(2, a), Ok(&"c"));
+    /// assert_eq!(cache.try_get_or_insert(3, a), Ok(&"d"));
+    /// assert_eq!(cache.try_get_or_insert(1, f), Err("failed".to_owned()));
+    /// assert_eq!(cache.try_get_or_insert(1, b), Ok(&"b"));
+    /// assert_eq!(cache.try_get_or_insert(1, a), Ok(&"b"));
+    /// ```
+    pub fn try_get_or_insert<F, E>(&mut self, k: K, f: F) -> Result<&V, E>
+    where
+        F: FnOnce() -> Result<V, E>,
+    {
+        if let Some(node) = self.map.get_mut(&KeyRef { k: &k }) {
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
+
+            self.detach(node_ptr);
+            self.attach(node_ptr);
+
+            unsafe { Ok(&*(*node_ptr).val.as_ptr()) }
+        } else {
+            match f() {
+                Err(e) => Err(e),
+                Ok(v) => {
+                    let (_, node) = self.replace_or_create_node(k, v);
+                    let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
+
+                    self.attach(node_ptr);
+
+                    let keyref = unsafe { (*node_ptr).key.as_ptr() };
+                    self.map.insert(KeyRef { k: keyref }, node);
+                    Ok(unsafe { &*(*node_ptr).val.as_ptr() })
+                }
+            }
         }
     }
 
@@ -537,22 +590,22 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
         F: FnOnce() -> V,
     {
         if let Some(node) = self.map.get_mut(&KeyRef { k: &k }) {
-            let node_ptr: *mut LruEntry<K, V> = &mut **node;
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
             self.detach(node_ptr);
             self.attach(node_ptr);
 
-            unsafe { &mut (*(*node_ptr).val.as_mut_ptr()) as &mut V }
+            unsafe { &mut *(*node_ptr).val.as_mut_ptr() }
         } else {
             let v = f();
-            let (_, mut node) = self.replace_or_create_node(k, v);
+            let (_, node) = self.replace_or_create_node(k, v);
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
 
-            let node_ptr: *mut LruEntry<K, V> = &mut *node;
             self.attach(node_ptr);
 
             let keyref = unsafe { (*node_ptr).key.as_ptr() };
             self.map.insert(KeyRef { k: keyref }, node);
-            unsafe { &mut (*(*node_ptr).val.as_mut_ptr()) as &mut V }
+            unsafe { &mut *(*node_ptr).val.as_mut_ptr() }
         }
     }
 
@@ -575,12 +628,12 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn peek<'a, Q>(&'a self, k: &Q) -> Option<&'a V>
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
         self.map
-            .get(k)
-            .map(|node| unsafe { &(*node.val.as_ptr()) as &V })
+            .get(KeyWrapper::from_ref(k))
+            .map(|node| unsafe { &*node.as_ref().val.as_ptr() })
     }
 
     /// Returns a mutable reference to the value corresponding to the key in the cache or `None`
@@ -602,12 +655,12 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn peek_mut<'a, Q>(&'a mut self, k: &Q) -> Option<&'a mut V>
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        match self.map.get_mut(k) {
+        match self.map.get_mut(KeyWrapper::from_ref(k)) {
             None => None,
-            Some(node) => Some(unsafe { &mut (*node.val.as_mut_ptr()) as &mut V }),
+            Some(node) => Some(unsafe { &mut *(*node.as_ptr()).val.as_mut_ptr() }),
         }
     }
 
@@ -662,10 +715,10 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn contains<Q>(&self, k: &Q) -> bool
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        self.map.contains_key(k)
+        self.map.contains_key(KeyWrapper::from_ref(k))
     }
 
     /// Removes and returns the value corresponding to the key from the cache or
@@ -687,18 +740,23 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn pop<Q>(&mut self, k: &Q) -> Option<V>
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        match self.map.remove(k) {
+        match self.map.remove(KeyWrapper::from_ref(k)) {
             None => None,
-            Some(mut old_node) => {
-                unsafe {
+            Some(old_node) => {
+                let mut old_node = unsafe {
+                    let mut old_node = *Box::from_raw(old_node.as_ptr());
                     ptr::drop_in_place(old_node.key.as_mut_ptr());
-                }
-                let node_ptr: *mut LruEntry<K, V> = &mut *old_node;
-                self.detach(node_ptr);
-                unsafe { Some(old_node.val.assume_init()) }
+
+                    old_node
+                };
+
+                self.detach(&mut old_node);
+
+                let LruEntry { key: _, val, .. } = old_node;
+                unsafe { Some(val.assume_init()) }
             }
         }
     }
@@ -724,15 +782,18 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn pop_entry<Q>(&mut self, k: &Q) -> Option<(K, V)>
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        match self.map.remove(k) {
+        match self.map.remove(KeyWrapper::from_ref(k)) {
             None => None,
-            Some(mut old_node) => {
-                let node_ptr: *mut LruEntry<K, V> = &mut *old_node;
-                self.detach(node_ptr);
-                unsafe { Some((old_node.key.assume_init(), old_node.val.assume_init())) }
+            Some(old_node) => {
+                let mut old_node = unsafe { *Box::from_raw(old_node.as_ptr()) };
+
+                self.detach(&mut old_node);
+
+                let LruEntry { key, val, .. } = old_node;
+                unsafe { Some((key.assume_init(), val.assume_init())) }
             }
         }
     }
@@ -789,11 +850,11 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn promote<'a, Q>(&'a mut self, k: &Q)
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        if let Some(node) = self.map.get_mut(k) {
-            let node_ptr: *mut LruEntry<K, V> = &mut **node;
+        if let Some(node) = self.map.get_mut(KeyWrapper::from_ref(k)) {
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
             self.detach(node_ptr);
             self.attach(node_ptr);
         }
@@ -825,11 +886,11 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
     /// ```
     pub fn demote<'a, Q>(&'a mut self, k: &Q)
     where
-        KeyRef<K>: Borrow<Q>,
+        K: Borrow<Q>,
         Q: Hash + Eq + ?Sized,
     {
-        if let Some(node) = self.map.get_mut(k) {
-            let node_ptr: *mut LruEntry<K, V> = &mut **node;
+        if let Some(node) = self.map.get_mut(KeyWrapper::from_ref(k)) {
+            let node_ptr: *mut LruEntry<K, V> = node.as_ptr();
             self.detach(node_ptr);
             self.attach_last(node_ptr);
         }
@@ -1018,10 +1079,10 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
             let old_key = KeyRef {
                 k: unsafe { &(*(*(*self.tail).prev).key.as_ptr()) },
             };
-            let mut old_node = self.map.remove(&old_key).unwrap();
-            let node_ptr: *mut LruEntry<K, V> = &mut *old_node;
+            let old_node = self.map.remove(&old_key).unwrap();
+            let node_ptr: *mut LruEntry<K, V> = old_node.as_ptr();
             self.detach(node_ptr);
-            Some(old_node)
+            unsafe { Some(Box::from_raw(node_ptr)) }
         } else {
             None
         }
@@ -1057,9 +1118,10 @@ impl<K: Hash + Eq, V, S: BuildHasher> LruCache<K, V, S> {
 
 impl<K, V, S> Drop for LruCache<K, V, S> {
     fn drop(&mut self) {
-        self.map.values_mut().for_each(|e| unsafe {
-            ptr::drop_in_place(e.key.as_mut_ptr());
-            ptr::drop_in_place(e.val.as_mut_ptr());
+        self.map.drain().for_each(|(_, node)| unsafe {
+            let mut node = *Box::from_raw(node.as_ptr());
+            ptr::drop_in_place((node).key.as_mut_ptr());
+            ptr::drop_in_place((node).val.as_mut_ptr());
         });
         // We rebox the head/tail, and because these are maybe-uninit
         // they do not have the absent k/v dropped.
@@ -1385,6 +1447,32 @@ mod tests {
         assert_eq!(cache.get_or_insert("lemon", || "red"), &"orange");
     }
 
+    #[test]
+    fn test_try_get_or_insert() {
+        let mut cache = LruCache::new(NonZeroUsize::new(2).unwrap());
+
+        assert_eq!(
+            cache.try_get_or_insert::<_, &str>("apple", || Ok("red")),
+            Ok(&"red")
+        );
+        assert_eq!(
+            cache.try_get_or_insert::<_, &str>("apple", || Err("failed")),
+            Ok(&"red")
+        );
+        assert_eq!(
+            cache.try_get_or_insert::<_, &str>("banana", || Ok("orange")),
+            Ok(&"orange")
+        );
+        assert_eq!(
+            cache.try_get_or_insert::<_, &str>("lemon", || Err("failed")),
+            Err("failed")
+        );
+        assert_eq!(
+            cache.try_get_or_insert::<_, &str>("banana", || Err("failed")),
+            Ok(&"orange")
+        );
+    }
+
     #[test]
     fn test_put_and_get_or_insert_mut() {
         let mut cache = LruCache::new(NonZeroUsize::new(2).unwrap());
